/* quote (a fortune replacement with rot13)
 * Copyright (c) 2002-2004, Emil Mikulic.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
 * IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * -=-
 *
 * Format is like BSD fortune, have a look at the bottom of this file.
 * If the first line reads "#rot13" fortunes will be deciphered before
 * being printed.
 *
 * Quotes are loaded from ~/.quotes by default.  This can be overridden
 * using the QUOTES environment variable.
 *
 */

static const char rcsid[] =
"$Id: quote.c,v 1.11 2004/11/16 05:45:29 emikulic Exp $";

#include <sys/types.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__linux__)
#include <err.h>
#else
#include <stdarg.h>
#include <errno.h>
static void warn(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", strerror(errno));
}
static void warnx(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");
}
static void errx(const int errcode, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, "\n");
	exit(errcode);
}
static void err(const int errcode, const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", strerror(errno));
	exit(errcode);
}
#endif

static char default_quotes_fn[] = /* $HOME/ */ ".quotes";
static char default_idx_suffix[] = ".idx";
static char quotes_env_var[] = "QUOTES";

static char rot13_header[] = "#rot13\n";
static char empty[] = "# No quotes file found.\n";
static char default_editor[] = "vi";

static char *quotes, *quotes_idx;

#ifdef __sun__
# define srand srand48
# define rand lrand48
#endif

#if __FreeBSD__
# define RANDOMIZE srandomdev()
# define RANDOM    (random())
#else
#if __OpenBSD__
# define RANDOMIZE do { } while (0)
# define RANDOM    (arc4random())
#else
# include <sys/time.h>
# define RANDOMIZE do { struct timeval t; gettimeofday(&t,NULL); \
                     srand(t.tv_sec + t.tv_usec + getpid()); } while (0)
# define RANDOM    (rand())
#endif
#endif



static void *xmalloc(const size_t s)
{
	void *tmp = malloc(s);

	if (tmp == NULL) errx(1, "out of memory");
	return tmp;
}



static time_t getmtime(const char *filename)
{
	struct stat s;

	if (stat(filename, &s) == -1)
		err(1, "stat %s", filename);
	return s.st_mtime;
}



static off_t getsize(const char *filename)
{
	struct stat s;

	if (stat(filename, &s) == -1)
		err(1, "stat %s", filename);
	return s.st_size;
}



static void init_paths(void)
{
	char *home;

	do {
		quotes = getenv(quotes_env_var);
		if (quotes != NULL)
			break;

		home = getenv("HOME");
		if (home == NULL)
			errx(1, "$HOME is not set");

		quotes = xmalloc(strlen(home) + 1 +
			sizeof(default_quotes_fn));
			/* sizeof includes \0 */
		sprintf(quotes, "%s/%s", home, default_quotes_fn);
	} while (0);

	quotes_idx = xmalloc(
		strlen(quotes) + sizeof(default_idx_suffix));
		/* sizeof includes \0 */
	sprintf(quotes_idx, "%s%s", quotes, default_idx_suffix);
}



#define rot13(c) ( (c >= 'N' && c <= 'Z')?(c - 'N' + 'A') : \
                   (c >= 'A' && c <= 'M')?(c - 'A' + 'N') : \
                   (c >= 'n' && c <= 'z')?(c - 'n' + 'a') : \
                   (c >= 'a' && c <= 'm')?(c - 'a' + 'n') : c)



static void do_usage(const char *me)
{
	printf(
		"usage: %s [-r|-c|-i|-e]\n"
		"\n"
		"\t-r\trot13 stdin to stdout\n"
		"\t-c\tcount number of quotes\n"
		"\t-i\t[re]index quotes\n"
		"\t-e\tedit quotes file\n"
		"\n", me);
}



/* stdin ---rot13---> stdout */
static void do_rot13(void)
{
	int c;

	while ( (c = getchar()) != EOF )
		putchar(rot13(c));
}



#define BUFSIZE 20
static char linebuf[BUFSIZE];
static void getline(FILE *f)
{
	int i;
	char c = 0;

	for (i=0; !feof(f) && c != '\n' && i < BUFSIZE-1; i++)
	{
		c = fgetc(f);
		linebuf[i] = c;
	}
	linebuf[i] = 0;

	/* read to end of line */
	while (!feof(f) && c != '\n') c = fgetc(f);
}
#undef BUFSIZE



static void xfwrite(const void *ptr, const size_t size, FILE *fp)
{
	if (fwrite(ptr, size, 1, fp) != 1)
		err(1, "writing to file");
}



/* generate quote index for faster quoting */
static void do_index(const int force)
{
	FILE *fq, *fi;
	unsigned int count, progress;
	long begin, end;
	unsigned char use_rot13 = 0;
	off_t size;

	/* we can skip indexing if it's not forced */
	if (!force)
	{
		struct stat is;
		if (stat(quotes_idx, &is) == 0)
		{
			/* index exists, don't rebuild if quotes are
			 * older than index.
			 */
			if (getmtime(quotes) < is.st_mtime) return;
		}
	}



	fq = fopen(quotes, "rb");
	fi = fopen(quotes_idx, "wb");

	if (!fq) err(1, "can't open %s for reading", quotes);
	if (!fi) err(1, "can't open %s for writing", quotes_idx);
	size = getsize(quotes);

	count = 0;
	begin = 0;
	getline(fq);
	if (strncmp(linebuf, rot13_header, sizeof(rot13_header)-2) == 0)
		use_rot13 = 1;

	xfwrite(&use_rot13, sizeof(use_rot13), fi);
	xfwrite(&count, sizeof(count), fi);

	/* eat up top-of-file comments */
	if (linebuf[0] == '#')
		do
		{
			begin = ftell(fq);
			getline(fq);
		}
		while (!feof(fq) && linebuf[0] == '#');

	/* first quote */
	do
	{
		end = ftell(fq);
		getline(fq);
	}
	while (!feof(fq) && strcmp(linebuf, "%\n") != 0);

	count = 1;
	xfwrite(&begin, sizeof(begin), fi);
	xfwrite(&end,   sizeof(end),   fi);

	progress = 0;
	fprintf(stderr, "      ]\r[");

	/* other quotes */
	while (!feof(fq))
	{
		unsigned int new_progress;
		begin = ftell(fq);
		new_progress = begin * 5 / size;
		for (; progress < new_progress; progress++)
			fprintf(stderr, ".");
		do
		{
			end = ftell(fq);
			getline(fq);
		}
		while (!feof(fq) && strcmp(linebuf, "%\n") != 0);

		if (end > begin)
		{
			count++;
			xfwrite(&begin, sizeof(begin), fi);
			xfwrite(&end,   sizeof(end),   fi);
		}
	}

	fprintf(stderr, "\r       \r");
	fseek(fi, (long)sizeof(use_rot13), SEEK_SET);
	xfwrite(&count, sizeof(count), fi);

	fclose(fi);
	fclose(fq);
}



static void do_count(void)
{
	FILE *f;
	unsigned int count;

	do_index(0);

	f = fopen(quotes_idx, "rb");
	if (f == NULL) err(1, "can't open %s", quotes_idx);

	fseek(f, (long)sizeof(char), SEEK_SET);
	if (fread(&count, sizeof(count), 1, f) != 1)
		err(1, "can't read from %s", quotes_idx);

	fclose(f);
	printf("There are %u quotes.\n", count);
}



static void do_quote(unsigned int num)
{
	int fd;
	int count, begin, end, size;
	unsigned char use_rot13;
	char *buf;

	do_index(0);
	fd = open(quotes_idx, O_RDONLY);
	if (fd == -1) err(1, "can't open %s", quotes_idx);
	(void)read(fd, &use_rot13, sizeof(use_rot13));
	(void)read(fd, &count, sizeof(count));

	/* pick a random quote */
	if (lseek(fd,
		(sizeof(begin)+sizeof(end)) * (num % count), SEEK_CUR) == -1)
		err(1, "lseek");
	(void)read(fd, &begin, sizeof(begin));
	(void)read(fd, &end,   sizeof(end));
	close(fd);
	size = end - begin;

	/* write out */
	buf = xmalloc(size);
	fd = open(quotes, O_RDONLY);
	if (fd == -1) err(1, "can't open %s", quotes);
	if (lseek(fd, begin, SEEK_SET) == -1)
		err(1, "lseek");
	(void)read(fd, buf, size);
	close(fd);

	if (use_rot13)
		for (count=0; count<size; count++)
			buf[count] = rot13(buf[count]);

	write(STDOUT_FILENO, buf, size);
	free(buf);
}



static void do_edit(void)
{
	char tmpname[18], *cmdline, *editor;
	int tmpfd, use_rot13 = 0;
	FILE *f, *tmpf;
	time_t mtime_before, mtime_after;

	/* make a temp file */
	strcpy(tmpname, "/tmp/quote.XXXXXX");
	tmpfd = mkstemp(tmpname);
	if (tmpfd == -1) err(1, "can't create temporary file");

	f = fopen(quotes, "r");
	if (f == NULL)
		(void) write(tmpfd, empty, sizeof(empty)-1);
	else
	{
		getline(f);
		if (strncmp(linebuf, rot13_header, sizeof(rot13_header)-2)
			== 0)
		{
			use_rot13 = 1;
			(void) write(tmpfd, rot13_header,
				sizeof(rot13_header)-1);
		}
		else
		{
			use_rot13 = 0;
			fseek(f, 0, SEEK_SET);
		}

		while (!feof(f))
		{
			#define BUFSIZE 1024
			char buf[BUFSIZE];
			size_t bufused;

			bufused = fread(buf, 1, BUFSIZE, f);
			if (use_rot13)
			{
				size_t i;
				for (i=0; i<bufused; i++)
					buf[i] = rot13(buf[i]);
			}
			(void) write(tmpfd, buf, bufused);
			#undef BUFSIZE
		}
		fclose(f);
	}
	close(tmpfd);

	/* run the editor */
	mtime_before = getmtime(tmpname);
	editor = getenv("EDITOR");
	if (editor == NULL || strlen(editor) == 0)
	    editor = default_editor;

	cmdline = xmalloc(strlen(editor)+1+strlen(tmpname)+1);
	sprintf(cmdline, "%s %s", editor, tmpname);
	if (system(cmdline) != EXIT_SUCCESS)
	{
		warn("failed to execute `%s'", cmdline);
		unlink(tmpname);
		exit(EXIT_FAILURE);
	}
	free(cmdline);

	/* bail out if not modified */
	mtime_after = getmtime(tmpname);
	if (mtime_before == mtime_after)
	{
		unlink(tmpname);
		warnx("%s not modified", tmpname);
		return;
	}

	/* put back into quotes file */
	tmpf = fopen(tmpname, "r");
	if (tmpf == NULL)
	{
		warn("can't open %s for reading", tmpname);
		unlink(tmpname);
		exit(EXIT_FAILURE);
	}

	f = fopen(quotes, "w");
	if (f == NULL)
	{
		warn("can't open %s for writing", quotes);
		unlink(tmpname);
		exit(EXIT_FAILURE);
	}

	getline(tmpf);
	if (strncmp(linebuf, rot13_header, sizeof(rot13_header)-2) == 0)
	{
		use_rot13 = 1;
		xfwrite(rot13_header, sizeof(rot13_header)-1, f);
	}
	else
	{
		use_rot13 = 0;
		fseek(tmpf, 0, SEEK_SET);
	}
		
	while (!feof(tmpf))
	{
		#define BUFSIZE 1024
		char buf[BUFSIZE];
		size_t bufused;

		bufused = fread(buf, 1, BUFSIZE, tmpf);
		if (use_rot13)
		{
			size_t i;
			for (i=0; i<bufused; i++)
				buf[i] = rot13(buf[i]);
		}
		xfwrite(buf, bufused, f);
		#undef BUFSIZE
	}
	fclose(tmpf);
	fclose(f);
	unlink(tmpname);

	printf("Quotes are in %s\n", quotes);
	printf("  (reindexing)\r");
	do_index(1);

	do_count();
}



int main(int argc, char **argv)
{
	init_paths();
	RANDOMIZE;

	if (argc > 2) do_usage(argv[0]); else
	if (argc == 2)
	{
		if (strcmp(argv[1], "-r") == 0) do_rot13(); else
		if (strcmp(argv[1], "-c") == 0) do_count(); else
		if (strcmp(argv[1], "-i") == 0) do_index(1); else
		if (strcmp(argv[1], "-e") == 0) do_edit(); else
		do_usage(argv[0]);
	}
	else
		do_quote(RANDOM);

	return EXIT_SUCCESS;
}



/*
#rot13
#
# Slap this in your ~/.quotes
#
Pybja pbageby gb Znb Mrqbat
Gnxr lbhe cebgrva cvyyf naq chg lbhe uryzrg ba
%
Pybja pbageby gb Znb Mrqbat
Pbzzrapvat pbhagqbja, ratvarf ba
*/
